type AsmVar a = a
type Dynamic a = a

infixl 1 :=, !:=
infixl 1 =>>
infixr 0  `seq`

-- Declaration of primitive functions ---------------------------------------

primitive primAsmEq     "primAsmEq"       :: a -> a -> Bool
primitive primAsmSteps  "primAsmSteps"    :: IO Int
primitive primAsmForceAssign "primAsmForceAssign" :: AsmVar a -> a -> IO ()
primitive primAsmSetHistory  "primAsmSetHistory"  :: Bool -> IO ()
primitive primAsmShowHistory "primAsmShowHistory" :: Int -> IO Bool
primitive primAsmMinHistory  "primAsmMinHistory"  :: IO Int
primitive primAsmVariable "primAsmNewFun0"  
    :: String -> Int -> (a -> a -> Bool) -> a -> AsmVar a
primitive primAsmFunction "primAsmNewFun1"  
    :: String -> Int -> val -> (key -> key -> Bool)
         -> (val -> val -> Bool) -> [(key,val)] -> AsmVar (key -> val)
primitive primAsmWrite  "primAsmWriteFun" :: AsmVar a -> a -> IO ()
primitive primAsmFire   "primAsmFire"     :: IO () -> IO Bool
primitive primAsmCall   "primAsmCall"     :: IO () -> IO ()
primitive primAsmSeq    "primAsmSeq"      :: IO () -> IO () -> IO ()
primitive primAsmCall1  "primAsmCall1"    :: IO () -> IO Bool
primitive primAsmSCall1 "primAsmSCall1"   :: IO () -> IO ()
primitive primAsmAssocs "primAsmAssocsFun":: AsmVar (key -> val) -> [(key,val)]
primitive primAsmRandom  "primAsmRandom"      :: Int -> Int
primitive primAsmDefault "primAsmDefault"     :: a
primitive primAsmInDom   "primAsmInDom"       :: key -> AsmVar (key -> val) -> Bool
primitive primAsmEmptyDom "primAsmEmptyDom"   :: AsmVar (key -> val) -> Bool
primitive primAsmStrictEval  "primAsmStrictEval" :: a -> IO a
primitive primAsmCopyExpr    "primAsmCopyExpr"   :: a -> a
primitive dynamicExpr        "primAsmCopyExpr"   :: a -> a
primitive primAsmNewInteger  "primAsmNewInteger" :: Int
primitive primAsmSetInteger  "primAsmSetInteger" :: Int -> IO ()

primitive primAsmDynFuns     "primAsmDynFuns"    :: IO [String]

primitive genericEq  "primGenericEq"  :: a -> a -> Bool
primitive genericCmp "primGenericCmp"  :: a -> a -> Int
primitive genericLt  "primGenericLt" :: a -> a -> Bool

primitive primTrace  "primTrace" :: String -> a -> a

-- Rule Monad ---------------------------------------------------------------

data Rule a = Rule (IO a)
            | Skip

skip :: Rule ()
skip = done

instance Functor Rule where
  map f (Rule m) = Rule (map f m)
  map f Skip     = Skip

instance Monad Rule where
  result = Rule  . result
  Rule m `bind` f = Rule (m `bind` (\x -> case f x of
                                           Rule r -> r
                                           Skip   -> result (error "skip occured")))
  Skip `bind` f   = f (error "skip does not return a value")


instance Monad0 Rule where
  zero = Skip

instance MonadPlus Rule where
  a ++ b = do a; b

(=>>) :: Monad m => Bool -> m a -> m ()
True  =>> act = void act
False =>> _   = done

-- AsmTerm Types --------------------------------------------------------------

class AsmOrd a where
   asmLt :: a -> a -> Bool

class AsmTerm a where
   (:=) :: AsmVar a -> a -> Rule ()
   l := r = Rule (primAsmWrite l r)

   asmDefault :: a
   asmDefault = primAsmDefault ()

(!:=) :: AsmVar a -> a -> IO ()
(!:=) = primAsmForceAssign

instance AsmTerm Bool
  where asmDefault = False

instance AsmTerm Int
instance AsmTerm Float
instance AsmTerm Char
instance AsmTerm String 

instance AsmTerm a => AsmTerm [a]
instance AsmTerm a => AsmTerm {a}
instance (AsmTerm a, AsmTerm b) => AsmTerm (a,b)
instance (AsmTerm a, AsmTerm b, AsmTerm c) => AsmTerm (a,b, c)
instance (AsmTerm a, AsmTerm b, AsmTerm c, AsmTerm d) => AsmTerm (a,b, c, d)

{-  
instance AsmTerm a => AsmTerm [a] where
  f := g  = seqs  (zipWith(:=) f g)

instance (AsmTerm a, AsmTerm b) => AsmTerm (a,b) where
   (l,r) := (l',r') = do l := l'
                         r := r'
-}


instance AsmTerm a => Eq a where
  (==) = genericEq

instance AsmTerm a => AsmOrd a where
  asmLt = genericLt

instance AsmTerm a => Text a where
  showsPrec i e = showString (primPrint i e [])

-- ASM dynamic variables and functions  ------------------------------------

initVal :: (Eq a, AsmTerm a) => String -> a -> AsmVar a
initVal name = primAsmVariable name 0 (==)

initValStdin :: String -> AsmVar String
initValStdin name = primAsmVariable name 1 (==) ""

initValStdout :: Bool -> String -> AsmVar String
initValStdout True name  = primAsmVariable name 3 (==) ""
initValStdout False name = primAsmVariable name 2 (==) ""

initAssocs :: (AsmTerm key, AsmTerm val, AsmOrd key, Eq val)
    => String -> [(key,val)] -> AsmVar (key -> val)
initAssocs name = primAsmFunction name 0 asmDefault asmLt (==)

initAssocsStdin :: (AsmTerm key, AsmOrd key)
    => String -> AsmVar (key -> String)
initAssocsStdin name = primAsmFunction name 1 asmDefault asmLt (==) []

initAssocsStdout :: (AsmTerm key, AsmOrd key)
    => Bool -> String -> AsmVar (key -> String)
initAssocsStdout True name = primAsmFunction name 3
                                  asmDefault asmLt (==) []
initAssocsStdout False name = primAsmFunction name 2
                                  asmDefault asmLt (==) []


initAssocs' :: (AsmTerm key, AsmTerm val, AsmOrd key, Eq val)
    => String -> val -> (key -> key -> Bool) -> (val -> val -> Bool) 
        -> [(key,val)] -> AsmVar (key -> val)
initAssocs' name dflt compare eq = primAsmFunction name 0 dflt compare eq

-- access functions --------------------------------------------------------

assocs :: AsmVar (key -> val) -> [(key,val)]
assocs = primAsmAssocs

dom :: Ord key => AsmVar (key -> val) -> {key}
dom = mkSet . map fst . assocs

ran :: Ord val => AsmVar (key -> val) -> {val}
ran = mkSet . map snd . assocs

inDom :: key -> AsmVar (key -> val) -> Bool
inDom = primAsmInDom

emptyDom :: AsmVar (key -> val) -> Bool
emptyDom = primAsmEmptyDom

notInDom :: key -> AsmVar (key -> val) -> Bool
notInDom key fun = not(key `inDom` fun)

dynFuns :: IO ()
dynFuns = do xs <- primAsmDynFuns
             seqs (map putStrLn xs)

-- combining rules ---------------------------------------------------------

forall :: [Rule ()] -> Rule ()
forall = seqs

rforall :: [Rule a] -> Rule [a]
rforall = binds

choose :: [Rule a] -> Rule ()
choose = \xs -> case xs of
                 [] -> Skip
                 xs -> do i <- Rule $ strictEval (primAsmRandom (length xs))
                          void (xs !! i)

chooseIfNone :: [Rule a] -> Rule () -> Rule ()
chooseIfNone = \xs r -> case xs of
                 [] -> r
                 xs -> do i <- Rule $ strictEval (primAsmRandom (length xs))
                          void (xs !! i)

choose_among :: [Rule ()] -> Rule ()
choose_among ([]) = skip
choose_among (xs) = do
    i <- Rule $ strictEval (primAsmRandom (length xs))
    let (ls,r:rs) = splitAt i xs
    b <- asmCall1 r
    if b then skip
         else choose_among (ls ++ rs)

iterate :: Rule () -> Rule ()
iterate = Rule . primAsmCall . dynamicExpr . rule2io

seq :: Rule () -> Rule () -> Rule ()
seq r1 r2 = Rule (primAsmSeq (dynamicExpr (rule2io r1))
                             (dynamicExpr (rule2io r2)))

secureCall :: Rule () -> Rule ()
secureCall = Rule . primAsmSCall1 . dynamicExpr . rule2io

asmCall1 :: Rule () -> Rule Bool
asmCall1 = Rule . primAsmCall1 . dynamicExpr . rule2io

data Agent = Agent Int

instance AsmTerm Agent
instance AsmTerm (Agent -> Rule ())

instance Eq (Agent -> Rule ()) where
  (==) = primAsmEq

instance Create Agent where
  createElem = Agent

agents :: AsmVar (Agent -> Bool)
agents = initAssocs "agents" []

act :: AsmVar (Agent -> (Agent -> Rule ()))
act = initAssocs "act" []


multi :: AsmVar (a -> (a -> Rule ())) -> Rule ()
multi = multi' . assocs
  where multi' [] = skip
        multi' as = do
          i <- Rule $ strictEval (primAsmRandom (length as))
          let (ls,(a,r):rs) = splitAt i as
          secureCall (r a)
          multi' (ls ++ rs)

-- general functions ------------------------------------------------------

rule2io :: Rule a -> IO ()
rule2io (Rule r) = void r
rule2io Skip     = done

rule2io' :: Rule a -> IO a
rule2io' (Rule r) = r
rule2io' Skip     = error "rule2io': skip not expected"

strictEval = primAsmStrictEval

activateHistory :: Bool -> IO ()
activateHistory = primAsmSetHistory

setDisplayState :: Int -> IO ()
setDisplayState i = do
  b <- primAsmShowHistory i
  if not b then putStrLn "no history information available"
           else done

stdout :: AsmVar String
stdout = initValStdout False "stdout"

stdin :: AsmVar String
stdin = initValStdin "stdin"



str2int :: String -> Int
str2int ('-':xs) = - (str2int xs)             
str2int xs = foldl (\r c -> 10*r + ord c - ord '0') 0 xs


-- fire functions on IO Monad ---------------------------------------------

fire1_ :: IO () -> IO Bool
fire1_ rule = primAsmFire (primAsmCopyExpr rule)


fireWhile_ :: Bool -> IO () -> IO ()
fireWhile_ f rule | primAsmCopyExpr f  
                      = do b1 <- fire1_ rule
                           if b1 then fireWhile_ f rule
                                 else done
fireWhile_ f rule | otherwise = done

fireUntil_ :: Bool -> IO () -> IO ()
fireUntil_ f rule = do b1 <- fire1_ rule
                       if b1 then if primAsmCopyExpr f 
                                   then done
                                   else fireUntil_ f rule
                             else done

fixpoint_ :: IO () -> IO ()
fixpoint_ r = do b <- primAsmFire (primAsmCopyExpr r)
                 if b then fixpoint_ r
                      else done

-- fire functions on Rule Monad --------------------------------------------

fire1 :: Rule a -> IO ()
fire1 = void . fire1_ . rule2io

fire :: Int -> Rule () -> IO ()
fire i = void . map or . binds . copy i . fire1_ . rule2io


fireWhile :: Bool -> Rule () -> IO ()
fireWhile f = fireWhile_ f . rule2io

fireUntil :: Bool -> Rule () -> IO ()
fireUntil f = fireUntil_ f .rule2io


fixpoint :: Rule () -> IO ()
fixpoint = fixpoint_ . rule2io


trace :: IO () -> Rule () -> Rule ()
trace f Skip     = Skip
trace f (Rule r) = Rule (do primAsmCopyExpr f; r)


newInteger :: Int
newInteger = primAsmNewInteger

newIntegers :: Int -> [Int]
newIntegers n = 
              take n
              . map (\i -> i + 0)
              . repeat
              . dynamicExpr $ newInteger + 0

asmSetInteger :: Int -> IO ()
asmSetInteger = primAsmSetInteger

-- create ----------------------------------------------------------------

class Create c where
  createElem    :: Int -> c

instance Create Int where
  createElem    = id

creates :: Create c => (Int, [c] -> Rule b) -> Rule b
creates (i,rule) = do
    let rs = map createElem (newIntegers i)
    rule (rs)

create :: Create c => Int -> ([c] -> Rule b) -> Rule b
create = \i rule -> creates(i,rule)

    
