INLINE Cell strict_eval(Cell e);
INLINE Void asmPrintString(FILE *out, Cell e);
extern Cell asmReadFromStdin();
extern Cell strict_eval2(Cell e);
extern Int  asmHistoryActive;
extern Int  asmHistoryLevel;
extern Int  asmHistoryStart;
extern Int  asmHistoryStop;

class LessCell {

 public:

  INLINE LessCell(Cell &ordFun) : m_ordFun(ordFun) { }

  INLINE bool operator()(const Cell a, const Cell b) {
    eval(ap(ap(m_ordFun,a),b));
    return (whnfHead==nameTrue);
  }

 private:
  
  Cell &m_ordFun;
};

class DynamicFunction;

class Value {
 public:
  INLINE Value() { }
  
  INLINE Value(int level, Cell value) 
    : m_level(level)
    , m_value(value) { }
  
  INLINE int level() { return m_level; }
  INLINE int value() { return m_value; }
  INLINE void setValue(Cell v) { m_value = v; }
  
 private:
  
  int  m_level;
  Cell m_value;
};

#if MSDEV
typedef std::list<Value> ValueList_t;
#else
typedef slist<Value> ValueList_t;
#endif

typedef struct AsmValue_t {
  ValueList_t oldValue;
  ValueList_t newValue;
} AsmValue_t;


class Update_t {
 public:
  INLINE Update_t(DynamicFunction *d, AsmValue_t *_v, Cell _arg)
    : m_dynFun(d)
    , m_v(_v)
    , m_arg(_arg) { }
  DynamicFunction *m_dynFun;
  AsmValue_t      *m_v;
  Cell            m_arg;
};


class DynamicFunction {

 public:

  typedef enum { Controlled = 0, Input = 1, 
		 Output = 2, OutputLoc = 3 } Kind_t;

  static int create(Cell funname, const Kind_t k, Cell eqFun, 
		    Cell defaultValue);
  static int create(Cell funname, const Kind_t k, Cell eqFun, 
		    Cell ordFun, Cell defaultValue, Cell initList);

  INLINE static DynamicFunction *check(Cell asmFun);
  INLINE Cell funname() const { return m_funname; }
  INLINE bool isDynamic0(){ return m_dynFun == Dynamic0; }
  INLINE bool isInput()   { return m_kind == Input; }
  INLINE bool outputLoc() { return m_kind == OutputLoc; }
  INLINE bool isOutput()  const { 
    return 
      (m_kind == Output) ||
      (m_kind == OutputLoc); 
  }
  INLINE bool supportsAssocs() const { 
    return m_dynFun == Dynamic1; 
  }
  INLINE Cell getAssocs();
  INLINE bool nonEmptyDom();
  INLINE bool inDom(Cell key);
  
  INLINE void markCells();
  INLINE void resetMark() { m_marked = false; }
  INLINE void setMark()   { m_marked = true; }
  INLINE bool isMarked()  { return m_marked; }

  ~DynamicFunction();

  INLINE static DynamicFunction *getDynFun(int id) {
    Globals_t &g = getGlobals();
    assert(id>=0 && id<g.size);
    assert(g.flist[id]);
    return g.flist[id];
  }
  static void initGarbageCollection();
  static void removeUnusedFunctions();

  INLINE Cell read();
  INLINE void write(Cell newValue, bool force);
  INLINE static bool updateState(bool makeOutput);

  INLINE static void openFrame() {
    Globals_t &g = getGlobals();
    g.level++;
    g.locationsOld = &g.locationMapOld[g.level];
    g.locationsNew = &g.locationMapNew[g.level];
  }

  INLINE static void openSecureFrame() {
    getGlobals().ignoreInconsistence++;
    openFrame();
  }
  INLINE static bool consistent() { 
    return getGlobals().isConsistent;
  }

  INLINE static int level() {
    return getGlobals().level;
  }

  INLINE void addUse() { m_useCounter++; }
  INLINE void removeUse() { m_useCounter--; }
  INLINE bool used() { return m_useCounter > 0; }


  static void closeFrame(bool writeSafe);
  static void closeSecureFrame();

  static void clearFrame();
  static void resetState();
  static void clearAll();
  static void initClass();

 private:

  DynamicFunction(Cell funname, const Kind_t k, Cell eqFun, 
		  Cell defaultValue);
  DynamicFunction(Cell funname, const Kind_t k, Cell eqFun, 
		  Cell ordFun, Cell defaultValue, Cell initList);
  INLINE int id() const { return m_id; }

  static int getNewID();
  INLINE static void markFunctionsToUpdate();

  typedef enum { Dynamic0, Dynamic1 } DynFun_t;

  typedef std::map<Cell, AsmValue_t, LessCell> Values_t;

#if MSDEV
  typedef std::list<Update_t> LocationList_t;
#else
  typedef slist<Update_t> LocationList_t;
#endif

  typedef struct Globals_t {
    DynamicFunction **flist;
    int size;
    int free;
    int level;
    int initCounter;
    int  ignoreInconsistence;
    bool isConsistent;
    std::map<int, LocationList_t> locationMapNew;
    std::map<int, LocationList_t> locationMapOld;
    LocationList_t *locationsNew;
    LocationList_t *locationsOld;
  } Globals_t;


  INLINE static Globals_t &getGlobals();


  INLINE void markLocation(AsmValue_t &value);
  INLINE void writeSafeLocation(AsmValue_t &v,
				Value newValue, 
				Cell  arg,
				bool init);
  INLINE void writeLocation(AsmValue_t *v,
			    ValueList_t &l, 
			    Value newValue,
			    Cell  arg,
			    bool init);
  INLINE void registerLocationNew(AsmValue_t *v, Cell arg);
  INLINE void registerLocationOld(AsmValue_t *v, Cell arg);
  static void outputValue(Update_t *u, bool takeNew);
  static void clearLevel(bool clearOld);
  INLINE bool eq(Cell a, Cell b);
  INLINE Cell readHistoryValue(ValueList_t &vl) {
    ValueList_t::iterator it;
    Cell v = ASMDEFAULT;
    for (it = vl.begin(); it != vl.end(); it++)
      if ((*it).level()<=asmHistoryLevel) {
	v = (*it).value();
	break;
      }
    return v;
  }
  
  Kind_t    m_kind;
  int       m_id;
  Globals_t &m_globals;
  Cell      m_eqFunction, m_ordFunction;
  Cell      m_defaultValue, m_initList;
  Cell      m_funname;
  bool      m_marked;
  int       m_useCounter;
  DynFun_t  m_dynFun;

  Values_t    *m_values;
  AsmValue_t  m_value;
};


Cell DynamicFunction::read() {
  AsmValue_t *value;
  Cell       arg = NIL;
  if (isOutput()) {
    ERROR(0) "reading output location not allowed" ETHEN
    ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
    ERRTEXT "\n"
    EEND; 
  }
  if (m_dynFun==Dynamic0) {
    if (whnfArgs!=0) { 
      ERROR(0) "no argument expected" ETHEN
      ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
      ERRTEXT "\n"
      EEND; 
    }
    value = &m_value;
  }
  else {
    if (whnfArgs!=1) { 
      ERROR(0) "exactly one argument expected" ETHEN
      ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
      ERRTEXT "\n"
      EEND; 
    }
    arg = strict_eval(pop());
    push(arg);
    Values_t::iterator it = m_values->find(arg);
    pop();
    last_evaluated_asmvar = funname();
    if (it == m_values->end()) {
      return m_defaultValue;
    }
    else value = &((*it).second);
  }
  if (isInput()) {
    Cell res;
    if (!value->newValue.empty() && !asmHistoryActive)
      return value->newValue.front().value();
    if (asmHistoryActive) {
      if (asmHistoryLevel<asmHistoryStop-asmHistoryStart) {
	// search
	ValueList_t::iterator it;
	for (it = value->oldValue.begin(); it != value->oldValue.end(); it++) {
	  if ((*it).level()==asmHistoryLevel)
	    return (*it).value();
	  if ((*it).level()<asmHistoryLevel) break;
	}
	return ASMDEFAULT;
      }
      if (!value->newValue.empty() &&
	  value->newValue.front().level()>=asmHistoryLevel)
	return value->newValue.front().value();
      if (!value->oldValue.empty() &&
	  value->oldValue.front().level()>=asmHistoryLevel)
	return value->oldValue.front().value();
    }
    // ask for value
    printf("\n");
    printExp(stdout,funname());
    if (m_values) {
      printf(" ");
      printExp(stdout,arg);
    }
    printf(" ? ");
    fflush(stdout);
    res = asmReadFromStdin();
    value->newValue.push_front(Value(m_globals.level,res));
    registerLocationNew(value,arg);
    return res;
  }
  else {
    last_evaluated_asmvar = funname();
    if (value->oldValue.empty()) {
      return m_defaultValue;
    }
    else {
      Cell v;
      if (!asmHistoryActive || 
	  (asmHistoryLevel==asmHistoryStop-asmHistoryStart))
	v = value->oldValue.front().value();
      else
        v = readHistoryValue(value->oldValue);
      if (v==ASMDEFAULT) return m_defaultValue;
      else return v;
    }
  }
}


INLINE Cell eval_value(Cell e)
{
  Cell res;
  sp_at_top = sp;
  res = strict_eval(e);
  sp_at_top = asm_SP;
  return res;
}


INLINE Cell strict_eval(Cell e) 
{
  Cell res;

  res = strict_eval2(e);
  while (isPair(res) && fst(res)==INDIRECT) res = snd(res);
  return res;
}



